#!/usr/bin/env python3

# Copyright (c) Free Software Foundation, Inc. All rights reserved.
# Licensed under the AGPL-3.0-only License. See LICENSE in the project root
# for license information.

import unittest
from unittest import mock

from fastapi.testclient import TestClient

from . import (
    URL,
    QUERY_ID,
    X_ROAD_INSTANCE,
    MEMBER_CLASS,
    MEMBER_CODE,
    SUBSYSTEM_CODE,
    ASICE_TYPE,
    ASIC_VERIFIER_RESPONSE,
    datetime_parser,
    mocked_requests_get
)
from asicverifier.restful_api import RestfulApi


class TestRestfulApi(unittest.TestCase):
    def __init__(self, methodName: str = 'runTest'):
        super().__init__(methodName)
        self.maxDiff = None

    @mock.patch('asicverifier.requests.get', side_effect=mocked_requests_get)
    def test_app(self, _):
        client: TestClient = TestClient(RestfulApi.app())
        self.assertEqual(client.get('/').status_code, 200)
        self.assertDictEqual(
            client.post(
                '/',
                params={'conf_refresh': True, 'type': ASICE_TYPE.value},
                json={
                    'securityServerUrl': URL,
                    'queryId': QUERY_ID,
                    'xRoadInstance': X_ROAD_INSTANCE,
                    'memberClass': MEMBER_CLASS,
                    'memberCode': MEMBER_CODE,
                    'subsystemCode': SUBSYSTEM_CODE
                }
            ).json(object_hook=datetime_parser),
            ASIC_VERIFIER_RESPONSE
        )

        with self.assertRaises(AttributeError):
            self.assertEqual(
                client.post(
                    '/',
                    params={'conf_refresh': True, 'type': ASICE_TYPE.value},
                    json={
                        'securityServerUrl': URL.replace('https', 'http'),
                        'queryId': QUERY_ID,
                        'xRoadInstance': X_ROAD_INSTANCE,
                        'memberClass': MEMBER_CLASS,
                        'memberCode': MEMBER_CODE,
                        'subsystemCode': SUBSYSTEM_CODE
                    }
                ).status_code
            )



#!/usr/bin/env python3

# Copyright (c) Free Software Foundation, Inc. All rights reserved.
# Licensed under the AGPL-3.0-only License. See LICENSE in the project root
# for license information.

from datetime import datetime
from enum import Enum
from io import BytesIO
import logging
import re
from os import getenv
from os.path import isdir
import subprocess
from tempfile import NamedTemporaryFile
from urllib.parse import urlencode, urljoin, urlparse
from zipfile import ZipFile

from dateutil import tz
from dateutil.parser import parse
from dotenv import load_dotenv
from importlib_metadata import PackageMetadata, metadata
import requests

load_dotenv()
META_DATA: PackageMetadata = metadata(__name__)
SUMMARY: str = META_DATA['Summary']


def to_datetime(string: str) -> datetime:
    return parse(string, tzinfos={'WIB': tz.gettz('Asia/Jakarta')})


def extract_subject_or_issuer(message: str) -> dict:
    return dict([
        element.split('=')
        for element in message.split(', ')
    ])


def extract_asice(message: str) -> dict:
    return {
        'verification': re.search(
            r'Verification (\w+)\.', message
        ).group(1),
        **{
            parent: {
                eldest_children: {
                    'subject': subject,
                    'issuer': issuer,
                    'serial_number': serial_number,
                    'valid': {'from': valid_from, 'until': valid_until}
                },
                **youngest_child
            }
            for (
                (parent, eldest_children),
                (subject, issuer, serial_number, valid_from, valid_until),
                youngest_child
            ) in zip(
                (
                    ('signer', 'certificate'),
                    ('ocsp_response', 'signed_by'),
                    ('timestamp', 'signed_by')
                ),
                zip(
                    map(
                        extract_subject_or_issuer,
                        re.findall(
                            r'Subject: (.+)\s+', message
                        )
                    ),
                    map(
                        extract_subject_or_issuer,
                        re.findall(
                            r'Issuer: (.+)\s+', message
                        )
                    ),
                    map(
                        int,
                        re.findall(
                            r'Serial number: (.+)\s+', message
                        )
                    ),
                    map(
                        to_datetime,
                        re.findall(
                            r'Valid from: (.+)\s+', message
                        )
                    ),
                    map(
                        to_datetime,
                        re.findall(
                            r'Valid until: (.+)\s+', message
                        )
                    )
                ),
                (
                    {
                        'id': {
                            key.lower(): value
                            for key, value in [
                                re.search(
                                    r'ID: (.+)\s+', message
                                ).group(1).split(':')
                            ]
                        }
                    },
                    {
                        'produced_at': to_datetime(
                            re.search(
                                r'Produced at: (.+)\s+', message
                            ).group(1)
                        )
                    },
                    {
                        'date': to_datetime(
                            re.search(r'Date: (.+)\s+', message).group(1)
                        )
                    }
                )
            )
        },
        'file': [
            {'path': path, 'digist': digist, 'status': status}
            for path, digist, status in re.findall(
                r'digest for \"(.+)\" is: (.+) \((.+)\)', message
            )
        ]
    }


class AsiceType(str, Enum):
    REQUEST: str = 'request'
    RESPONSE: str = 'response'


def asicverifier(
    security_server_url: str,
    query_id: str,
    x_road_instance: str,
    member_class: str,
    member_code: str,
    subsystem_code: str,
    asice_type: AsiceType = AsiceType.REQUEST,
    conf_refresh: bool = False
) -> dict:
    CONF_PATH: str = (
        f'asicverifier/security-server/{urlparse(security_server_url).netloc}/'
    )

    if conf_refresh or not isdir(CONF_PATH):
        # Zip File
        response = requests.get(
            urljoin(security_server_url, 'verificationconf'),
            allow_redirects=True
        )

        try:
            response.raise_for_status()

            with ZipFile(BytesIO(response.content)) as zip_file:
                zip_file.extractall(CONF_PATH)
        except requests.exceptions.HTTPError:
            logging.error(
                '\n'.join([
                    f'{key}: "{value}"'
                    for key, value in dict(
                        security_server_url=security_server_url,
                        verificationconf=response.text
                    ).items()
                ])
            )
            raise

    # Asice File
    response = requests.get(
        '{url}?unique&{asice_type}&{params}'.format(
            url=urljoin(security_server_url, 'asic'),
            asice_type={
                enum.value: f'{enum.value}Only'
                for enum in AsiceType
            }[asice_type.value],
            params=urlencode({
                'queryId': query_id,
                'xRoadInstance': x_road_instance,
                'memberClass': member_class,
                'memberCode': member_code,
                'subsystemCode': subsystem_code
            })
        ),
        allow_redirects=True
    )

    try:
        response.raise_for_status()

        with NamedTemporaryFile() as temp, subprocess.Popen(
                ['yes', 'n'], stdout=subprocess.PIPE) as proc_yes:
            temp.write(response.content)
            temp.seek(0)
            message: str = subprocess.run(
                [
                    'java', '-jar',
                    getenv('JAR_PATH', '/lib/asicverifier.jar'),
                    f'{CONF_PATH}/verificationconf/',
                    temp.name
                ],
                stdin=proc_yes.stdout,
                capture_output=True
            ).stdout.decode()
            proc_yes.kill()
            logging.debug(message)
    except requests.exceptions.HTTPError:
        logging.error(
            '\n'.join([
                f'{key}: "{value}"'
                for key, value in dict(
                    security_server_url=security_server_url,
                    query_id=query_id,
                    x_road_instance=x_road_instance,
                    member_class=member_class,
                    member_code=member_code,
                    subsystem_code=subsystem_code,
                    asice_type=asice_type,
                    asice=response.text
                ).items()
            ])
        )
        raise

    return extract_asice(message)



#!/usr/bin/env python3

# Copyright (c) Free Software Foundation, Inc. All rights reserved.
# Licensed under the AGPL-3.0-only License. See LICENSE in the project root
# for license information.

from datetime import datetime
from os import getenv
from typing import List

from fastapi import APIRouter, FastAPI, HTTPException, Query
from fastapi.middleware.cors import CORSMiddleware
from pydantic import (
    BaseModel,
    EmailStr,
    Field,
    HttpUrl,
    NonNegativeInt
)
from requests.exceptions import HTTPError
import uvicorn

from . import AsiceType, asicverifier, META_DATA, SUMMARY


class AsicSignerId(BaseModel):
    subsystem: str


class AsicSignValid(BaseModel):
    from_: datetime = Field(alias='from')
    until: datetime


class AsicOcsp(BaseModel):
    CN: str
    O: str


class AsicSign(BaseModel):
    subject: AsicOcsp
    issuer: AsicOcsp
    serial_number: NonNegativeInt
    valid: AsicSignValid


class AsicSignerCertificateSubject(AsicOcsp):
    C: str = None


class AsicSignerCertificate(AsicSign):
    subject: AsicSignerCertificateSubject


class AsicSigner(BaseModel):
    certificate: AsicSignerCertificate
    id: AsicSignerId


class AsicOcspResponse(BaseModel):
    signed_by: AsicSign
    produced_at: datetime


class AsicTimeStampSignByIssuer(AsicSignerCertificateSubject, BaseModel):
    ST: str = None
    L: str = None
    EMAILADDRESS: EmailStr = None
    OU: str = None


class AsicTimeStampSignBySubject(AsicTimeStampSignByIssuer):
    oid_2_5_4_13: str = Field(None, alias='OID.2.5.4.13')


class AsicTimeStampSignBy(AsicSign):
    subject: AsicTimeStampSignBySubject
    issuer: AsicTimeStampSignByIssuer


class AsicTimeStamp(BaseModel):
    signed_by: AsicTimeStampSignBy
    date: datetime


class AsicFile(BaseModel):
    path: str
    digist: str
    status: str


class Asice(BaseModel):
    verification: str
    signer: AsicSigner
    ocsp_response: AsicOcspResponse
    timestamp: AsicTimeStamp
    file: List[AsicFile]


non_empty_str: str = r'^[\w\-]+$'


class AsicVerifier(BaseModel):
    security_server_url: HttpUrl = Field(alias='securityServerUrl')
    query_id: str = Field(alias='queryId', pattern=non_empty_str)
    x_road_instance: str = Field(alias='xRoadInstance', pattern=non_empty_str)
    member_class: str = Field(alias='memberClass', pattern=non_empty_str)
    member_code: str = Field(alias='memberCode', pattern=non_empty_str)
    subsystem_code: str = Field(alias='subsystemCode', pattern=non_empty_str)


class RestfulApi:
    @staticmethod
    def app() -> FastAPI:
        RESTFUL_API_PATH: str = getenv('RESTFUL_API_PATH', '/')

        if RESTFUL_API_PATH.endswith('/'):
            RESTFUL_API_PATH = RESTFUL_API_PATH[:-1]

        api: FastAPI = FastAPI(
            title=SUMMARY,
            version=META_DATA['Version'],
            contact=dict(
                name=META_DATA['Author'],
                url=META_DATA['Home-page'],
                email=META_DATA['Author-email']
            ),
            license_info=dict(
                name=META_DATA['License'],
                identifier=META_DATA['License'],
                url=f"{META_DATA['Home-page']}/blob/main/LICENSE"
            ),
            docs_url=f'{RESTFUL_API_PATH}/',
            redoc_url=f'{RESTFUL_API_PATH}/redoc',
            openapi_url=f'{RESTFUL_API_PATH}/openapi.json'
        )
        api.add_middleware(
            CORSMiddleware,
            allow_origins=[
                'http://0.0.0.0',
                'http://localhost',
                'http://localhost:8080'
            ],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*']
        )
        router = APIRouter()

        @router.post('/')
        async def verifier(
            data: AsicVerifier,
            asice_type: AsiceType = Query(
                None, alias='type', description='Default is request'
            ),
            conf_refresh: bool = Query(None, description='Default is false')
        ) -> Asice:
            try:
                return asicverifier(
                    **{key: f'{value}' for key, value in data},
                    asice_type=asice_type if asice_type else AsiceType.REQUEST,
                    conf_refresh=conf_refresh
                )
            except HTTPError as error:
                raise HTTPException(error.response.status_code)

        api.include_router(router, prefix=RESTFUL_API_PATH)
        return api

    @staticmethod
    def run(
        host: str = '0.0.0.0', port: int = 80, reload: bool = False
    ):
        'RESTful API'

        uvicorn.run(
            f'{__name__}:RestfulApi.app',
            host=host,
            port=port,
            reload=reload,
            factory=True
        )  # pragma: no cover
